/* Strip Club - Online/Offline Comic Reader/Archiver
 *
 * Copyright notice for this file:
 *  Copyright (C) 2004 Benjamin Cutler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <SDL/SDL_net.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include "interface.h"
#include "http.h"
#include "url.h"
#include "load.h"

IPaddress ip;	// Holds host info for current comic, set in SwitchComic()

bool GetRemoteInfo(const sURL *URL, const char *Referer, time_t *ModTime, int *ContentLength, char *ContentType) {
	TCPsocket socket;
	SDLNet_SocketSet set;	// For polling if the connection is slow
	char Label[80] = "";	// Bar Label
	char Header[80];
	char Buffer[400];
	char Request[2048];
	char month[4];
	char temp;
	int Cur = 0, timeout = 0;
	tm time;
	sURL RefererURL;

	if (Referer) {
		URLBreak(Referer, &RefererURL);
	} else {
		// Note, make sure not to delete BOTH in this case, because the pointers contained are the same memory...
		memcpy(&RefererURL, URL, sizeof(sURL));		
	}

	if (ModTime) {
		*ModTime = 0;
	}
	if (ContentLength) {
		*ContentLength = 0;
	}

	if(strcmp(URL->Host, RefererURL.Host)) {	// Look up host again...
		Interface.BarLabel("Looking up host...");
		if (SDLNet_ResolveHost(&ip, RefererURL.Host, atoi(RefererURL.Port))) {
			Interface.BarLabel("Unable to resolve host!");
			MOut("Unable to resolve host: %s\n", RefererURL.Host);
			URLDelete(&RefererURL);
			SDLNet_ResolveHost(&ip, URL->Host, atoi(URL->Port));	// Reset host
			return false;
		}
	}

	Interface.BarLabel("Opening connection...");

	socket = SDLNet_TCP_Open(&ip);

	if (!socket) {
		MOut("SDLNet_TCP_Open: %s\n", SDLNet_GetError());
		if(strcmp(URL->Host, RefererURL.Host)) {	// Look up host again...
			URLDelete(&RefererURL);
			SDLNet_ResolveHost(&ip, URL->Host, atoi(URL->Port));	// Reset host
		}
		return false;
	}

	set = SDLNet_AllocSocketSet(1);

	if (!set) {
		Out("SDLNet_AllocSocketSet: %s\n", SDLNet_GetError());
		SDLNet_TCP_Close(socket);
		if(strcmp(URL->Host, RefererURL.Host)) {	// Look up host again...
			URLDelete(&RefererURL);
			SDLNet_ResolveHost(&ip, URL->Host, atoi(URL->Port));	// Reset host
		}
		return false;
	}

	SDLNet_TCP_AddSocket(set, socket);

	DbgOut("Requesting HEAD for URL: %s\n", URL->Full);

	sprintf(Request, "HEAD %s HTTP/1.1\r\n", URL->FullPath);

	strcat(Request, "Host: ");	strcat(Request, URL->Host);	strcat(Request, "\r\n");

	if (Referer) {
		strcat(Request, "Referer: ");
		strcat(Request, Referer);
		strcat(Request, "\r\n");
	}

	strcat(Request, "User-Agent: ");
	if (CurComicG) {
		if (CurComicG->entryExists("useragent")) {
			char UserAgent[80];
			CurComicG->get("useragent", UserAgent, "", 79);
			strcat(Request, UserAgent);
		} else {
			strcat(Request, USERAGENT);
		}
	} else {
		strcat(Request, USERAGENT);
	}
	strcat(Request, "\r\n");
	strcat(Request, "Accept: text/html,text/plain;q=0.9,image/jpeg,image/png,image/gif;q=0.8\r\n");
	strcat(Request, "Accept-Language: en-us,en;q=0.5\r\n");
	strcat(Request, "Accept-Charset: ISO-8859-1;q=0.7\r\n");
	strcat(Request, "Accept-Encoding: \r\n");
	strcat(Request, "Connection: close\r\n");
	strcat(Request, "\r\n");

	UOut("Sending HTTP request to: %s (%X)\n", URL->Host, ip.host);
	UOut(Request);

	Interface.BarLabel("Sending HEAD Request...");

	SDLNet_TCP_Send(socket, (void *)Request, strlen(Request));

	Interface.BarLabel("Awaiting HEAD Response...");

	while(!SDLNet_CheckSockets(set, 100)) { 	// Wait for 1/10th second
		sprintf(Label, "Awaiting HEAD Response... (%ds)", (++timeout)/10);
//		DbgOut("Timeout: %d\n", timeout);
		Interface.BarLabel(Label);
	}

	SDLNet_TCP_DelSocket(set, socket);
	SDLNet_FreeSocketSet(set);

	if(SDLNet_TCP_Recv(socket, Header, 17) < 17) {
		MOut("Error retrieving HEAD, connection refused/closed?\n");
		SDLNet_TCP_Close(socket);
		if(strcmp(URL->Host, RefererURL.Host)) {	// Look up host again...
			URLDelete(&RefererURL);
			SDLNet_ResolveHost(&ip, URL->Host, atoi(URL->Port));	// Reset host
		}
		return false;
	} else if (strncmp(Header, "HTTP/1.1 200 OK", 15)) {
		int i = 17;
		do {
			SDLNet_TCP_Recv(socket, &temp, 1);
			Header[i++] = temp;
		} while((temp != 0xD) && (temp != 0xA));
		Header[i - 1] = '\0';
		MOut("Error retrieving HEAD, expected \"HTTP/1.1 200 OK\" but got \"%s\"\n", Header);
		SDLNet_TCP_Close(socket);
		if(strcmp(URL->Host, RefererURL.Host)) {	// Look up host again...
			URLDelete(&RefererURL);
			SDLNet_ResolveHost(&ip, URL->Host, atoi(URL->Port));	// Reset host
		}
		return false;
	}

	Interface.BarLabel("Receiving HEAD response...");

	Header[17] = '\0';
	DOut("Receiving HEAD response from: %s\n", URL->Host);
	DOut(Header);

	while (SDLNet_TCP_Recv(socket, &temp, 1)) {
		Buffer[Cur++] = temp;
		DOut("%c", temp);
		if (temp == 0xA) {
			if (Cur == 2) { 	// Done
				break;
			}
			sscanf(Buffer, "%s", Header);
			Cur = 0;
			if (!strcmp(Header, "Last-Modified:")) {

				if (ModTime) {
					sscanf((Buffer + strlen(Header)), "%*s %d %3s %d %d:%d:%d", &time.tm_mday, month, &time.tm_year, &time.tm_hour, &time.tm_min, &time.tm_sec);

					time.tm_year -= 1900;		// Struct is expected relative to 1900, but header returns full year
					if (!strcmp(month, "Jan")) {
						time.tm_mon = 0;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Feb")) {
						time.tm_mon = 1;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Mar")) {
						time.tm_mon = 2;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Apr")) {
						time.tm_mon = 3;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "May")) {
						time.tm_mon = 4;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Jun")) {
						time.tm_mon = 5;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Jul")) {
						time.tm_mon = 6;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Aug")) {
						time.tm_mon = 7;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Sep")) {
						time.tm_mon = 8;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Oct")) {
						time.tm_mon = 9;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Nov")) {
						time.tm_mon = 10;
						*ModTime = mktime(&time);
					} else if (!strcmp(month, "Dec")) {
						time.tm_mon = 11;
						*ModTime = mktime(&time);
					} else {
						Out("Invalid Month: %s\n", month);
						*ModTime = 0;
					}	// end Month Block
				}	// if (Modtime)
			} else if (!strcmp(Header, "Content-Length:")) {
				if (ContentLength) {
					sscanf((Buffer + strlen(Header)), "%d", ContentLength);
				}
			} else if (!strcmp(Header, "Content-Type:")) {
				if (ContentType) {
					sscanf(Buffer, "%*s %100[a-zA-Z1-9/]", ContentType);
					DbgOut("ContentType = %s\n",ContentType);
				}
			}
		}	// if (temp == 0xA) (Newline)
	}

	SDLNet_TCP_Close(socket);

	if(strcmp(URL->Host, RefererURL.Host)) {	// Look up host again...
		URLDelete(&RefererURL);
		SDLNet_ResolveHost(&ip, URL->Host, atoi(URL->Port));	// Reset host
	}

	return true;
}

char *GrabHTTPLink(const char *URL, const char *Referer) {
	TCPsocket socket;
	SDLNet_SocketSet set;
	char Buffer[4096];
	char Request[2048];
	static char File[2048];
	sURL Break;
	char temp;
	char Label[400];
	char type[100] = "";			// Content-Type
	int filelength, length, done = 0, Cur = 0, timeout = 0;
	time_t localmod, remotemod;
	struct stat filestat;
	char *Return;		// KEEP THIS, otherwise it will try to parse images for links
	FILE *fptr;
	bool body = false, chunked = false;		// Chunked = Chunked Transfer encoding?

	Interface.BarLabel("");
	Interface.BarRange(0.0f, 0.0f);
	Interface.BarValue(0.0f);

	URLBreak(URL, &Break);

	if(!GetRemoteInfo(&Break, Referer, &remotemod, &filelength, type)) {
		Interface.BarLabel("Unable to retrieve HEAD!");
		Out("Unable to retrieve HEAD for \"%s\"!\n", URL);
		Fl::wait(4);
		Interface.BarLabel("");
		URLDelete(&Break);
		return NULL;
	}

	if (!strcmp(type, "image/jpeg") || !strcmp(type, "image/gif") || !strcmp(type, "image/png")) {
		strcpy(File, GetCacheFileName(Break.File));
		Out("Local Filename: %s\n", File);
		if (!stat(File, &filestat)) {		// File exists
			localmod = filestat.st_mtime + timezone;
			DbgOut("Time: Old: %ld %s New: %ld\n", localmod, (localmod >= remotemod) ? ">=" : "< ", remotemod);
			DbgOut("Size: Old: %ld %s New: %ld\n", filestat.st_size, (filestat.st_size == filelength) ? "==" : "!=", filelength);
			if ((localmod >= remotemod) && (remotemod) && (filelength == filestat.st_size)) {		// File is newer or just as recent as the local copy, and remotemod != NULL, and filesizes match
				Interface.BarLabel("Local copy up to date...");
				Out("Local copy up to date\n");
				URLDelete(&Break);
				return File;
			}
		}
		if (!(fptr = fopen(File, "wb"))) {		// Rare, but hey
			Interface.BarLabel("Error opening output file!");
			Out("Error opening \"%s\" for writing!\n", File);
			Fl::wait(4);
			Interface.BarLabel("");
			return NULL;
		}
		Return = File;
	} else if (!strcmp(Break.Extension, "comic")) {
		strcpy(File, GetComicFileName(Break.File));
		if (!(fptr = fopen(File, "wb"))) {		// Rare, but hey
			Interface.BarLabel("Error opening output file!");
			Out("Error opening \"%s\" for writing!\n", File);
			Fl::wait(4);
			Interface.BarLabel("");
			return NULL;
		}
		Return = File;
	} else if (!strcmp(type, "text/plain")) {
		fptr = fopen("temp.txt", "wb");
		Return = "temp.txt";
	} else if (!strncmp(type, "text/html", 9)) {
		fptr = fopen("temp.html", "wb");
		Return = "temp.html";
	} else {		// What the hell are we pointing at?
		return NULL;
	}
	DbgOut("GrabHTTPLink() Return Value: %s\n", Return);

	Interface.BarLabel("Opening connection...");

	socket = SDLNet_TCP_Open(&ip);

	if (!socket) {
		MOut("SDLNet_TCP_Open: %s\n", SDLNet_GetError());
		return NULL;
	}

	set = SDLNet_AllocSocketSet(1);

	if (!set) {
		Out("SDLNet_AllocSocketSet: %s\n", SDLNet_GetError());
		SDLNet_TCP_Close(socket);
		return NULL;
	}

	SDLNet_TCP_AddSocket(set, socket);

	Interface.BarLabel("Sending GET Request...");

	Out("Requesting URL: %s\n", Break.Full);

	sprintf(Request, "GET %s HTTP/1.1\r\n", Break.FullPath);

	if (Referer) {
		strcat(Request, "Referer: ");
		strcat(Request, Referer);
		strcat(Request, "\r\n");
	}

	strcat(Request, "User-Agent: ");
	if (CurComicG) {
		if (CurComicG->entryExists("useragent")) {
			char UserAgent[80];
			CurComicG->get("useragent", UserAgent, "", 79);
			strcat(Request, UserAgent);
		} else {
			strcat(Request, USERAGENT);
		}
	} else {
		strcat(Request, USERAGENT);
	}
	strcat(Request, "\r\n");
	strcat(Request, "Accept: text/html,text/plain;q=0.9,image/jpeg,image/png,image/gif;q=0.8\r\n");
	strcat(Request, "Accept-Language: en-us,en;q=0.5\r\n");
	strcat(Request, "Accept-Charset: ISO-8859-1;q=0.7\r\n");
	strcat(Request, "Accept-Encoding: \r\n");
	strcat(Request, "Host: ");	strcat(Request, Break.Host);	strcat(Request, "\r\n");
	strcat(Request, "Connection: close\r\n");
	strcat(Request, "\r\n");

	UOut("Sending HTTP request to: %s\n", Break.Host);
	UOut(Request);

	SDLNet_TCP_Send(socket, (void *)Request, strlen(Request));

	Interface.BarLabel("Awaiting GET Response...");

	while(!SDLNet_CheckSockets(set, 100)) {
		sprintf(Label, "Awaiting GET Response... (%ds)", (++timeout)/10);
//		DbgOut("Timeout: %d\n", timeout);
		Interface.BarLabel(Label);
	}

	SDLNet_TCP_DelSocket(set, socket);
	SDLNet_FreeSocketSet(set);

	set = NULL;

	Cur = 0;

	DOut("Receiving GET response from: %s\n", Break.Host);
	Interface.BarLabel("Receiving GET Response...");

	while (!body) {
		if (SDLNet_TCP_Recv(socket, &temp, 1) <= 0) {
			MOut("Error retrieving URL: %s\n", SDLNet_GetError());
			SDLNet_TCP_Close(socket);
			URLDelete(&Break);
			return NULL;
		}
		Buffer[Cur++] = temp;
		DOut("%c", temp);
		if (temp == 0xA) {
			if (Cur == 2) { 
				body = true; 
			} else {
				char Header[80], Encoding[80];
				sscanf(Buffer, "%79s %79s", Header, Encoding);
				if (!strcmp(Header, "Transfer-Encoding:") && !strcmp(Encoding, "chunked")) { // "Transfer-Encoding: chunked"
					chunked = true;
				}
			}
			Cur = 0;
		}
	};

	DbgOut("Now receiving body\n");

	if (chunked) {
		DbgOut("Chunked transfer\n");
	}

	strcpy(Label, "");

	Interface.BarLabel(Label);

	if (filelength) {
		Interface.BarRange(0.0f, (float)filelength);
		DbgOut("File Length: %ld\n", filelength);
		if (!chunked) {
			do {
				length = SDLNet_TCP_Recv(socket, Buffer, 4096);
				if (length > 0) {
					fwrite(Buffer, 1, length, fptr);
					done += length;
					sprintf(Label, "%s %d/%d (%d%%)", strlen(Break.File) ? Break.File : "index.html", done, filelength, done*100/filelength);
					Interface.BarValue((float)done);
				}
			} while ((done < filelength) && (length > 0));
		} else {
			int chunklength, chunkdone;
			char chunkheader[16], tmp;
			do {
				do {
					length = SDLNet_TCP_Recv(socket, &tmp, 1);
					if (length < 1) {		// Connection broken
						MOut("Error: Connection closed before done receiving body\n");
						break;
					}
					chunkheader[Cur++] = tmp;
				} while (tmp != 0xA);
				chunkheader[Cur] = 0; 		// End string
				Cur = 0;
				sscanf(chunkheader, "%x", &chunklength);
				chunkdone = 0;
				DbgOut("Chunkheader: %s", chunkheader);		// Because the chunkheader has it's own newline
				DbgOut("Chunklength: %d", chunklength);
				if (!chunklength) {		// Hit the last chunk
					done = length;
					SDLNet_TCP_Close(socket);
					socket = NULL;
					DbgOut("Closed Socket\n");
				}
				while (chunkdone < chunklength) {
					length = SDLNet_TCP_Recv(socket, Buffer, min(chunklength - chunkdone, 4096));
					if (length < 1) {		// Connection broken
						MOut("Error: Connection closed before done receiving body\n");
						break;
					}
					chunkdone += length;
					done += length;
					sprintf(Label, "%s %d/%d (%d%%)", strlen(Break.File) ? Break.File : "index.html", done, filelength, done*100/filelength);
					Interface.BarValue((float)done);
					fwrite(Buffer, 1, length, fptr);
				}
				if (socket) {
					SDLNet_TCP_Recv(socket, Buffer, 2);	// Read the CRLF an the end of the chunk
				}
			} while ((done < filelength) && (length > 0));
		}
		Interface.BarValue(0.0f);
	} else {
		if (!chunked) {
			do {
				length = SDLNet_TCP_Recv(socket, Buffer, 4096);
				if (length > 0) {
						fwrite(Buffer, 1, length, fptr);
						done += length;
						sprintf(Label, "%s %d", strlen(Break.File) ? Break.File : "index.html", done);
						Interface.BarLabel(Label);	// To make sure the display is updated
				}
			} while (length > 0);
		} else {
			bool chunkalldone = false;
			int chunklength, chunkdone;
			char chunkheader[16], tmp;
			do {
				do {
					length = SDLNet_TCP_Recv(socket, &tmp, 1);
					if (length < 1) {		// Connection broken
						MOut("Error: Connection closed before done receiving body\n");
						break;
					}
					chunkheader[Cur++] = tmp;
				} while (tmp != 0xA);
				chunkheader[Cur] = 0; 		// End string
				Cur = 0;
				sscanf(chunkheader, "%x", &chunklength);
				chunkdone = 0;
				DbgOut("Chunkheader: %s", chunkheader);		// Because the chunkheader has it's own newline
				DbgOut("Chunklength: %d\n", chunklength);
				if (!chunklength) {		// Hit the last chunk
					chunkalldone = true;
					SDLNet_TCP_Close(socket);
					socket = NULL;
					DbgOut("Closed Socket\n");
				}
				while (chunkdone < chunklength) {
					length = SDLNet_TCP_Recv(socket, Buffer, min(chunklength - chunkdone, 4096));
					if (length < 1) {		// Connection broken
						MOut("Error: Connection closed before done receiving body\n");
						break;
					}
					done += length;
					chunkdone += length;
					sprintf(Label, "%s %d", strlen(Break.File) ? Break.File : "index.html", done);
					Interface.BarLabel(Label);	// To make sure the display is updated
					fwrite(Buffer, 1, length, fptr);
				}
				if (socket) {
					SDLNet_TCP_Recv(socket, Buffer, 2);	// Read the CRLF an the end of the chunk
				}
			} while ((!chunkalldone) && (length > 0));
		}
	}

	if (length < 0) {	// Rare, but just in case...
		MOut("SDLNet_TCP_Recv: %s\n", SDLNet_GetError());
	}

	fclose(fptr);

	Interface.BarLabel("");

	Out("Done with Request\n");

	if (socket) {
		SDLNet_TCP_Close(socket);
		socket = NULL;
		DbgOut("Closed Socket\n");
	}

	URLDelete(&Break);

	DbgOut("Returning: %s\n", Return);

	return Return;
}
