/* Strip Club - Online/Offline Comic Reader/Archiver
 *
 * Copyright notice for this file:
 *  Copyright (C) 2004 Benjamin Cutler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <sys/stat.h>
#include <unistd.h>
#endif
#include <pcre.h>
#include <SDL/SDL_net.h>
#include <FL/Fl_Box.H>
#include <FL/Fl_Preferences.H>
#include <FL/Fl_Text_Buffer.H>
#include "load.h"
#include "url.h"
#include "http.h"
#include "html.h"
#include "interface.cxx"

char *CurComic;
Fl_Preferences *Prefs = NULL, *CurComicG = NULL;
cWindow Interface;
PageLinks CurLinks;
eOffType Offline = OFFTYPE_CHECK;
char *OffArray[] = { "Offline", "Missing", "Check" };

void EditComic(Fl_Widget *w, void *Comic);

// import.cpp
void ExportComic(Fl_Widget *w, void *Comic);
void UpdateComics();

int main(int argc, char **argv) {
	Fl_Text_Buffer TextOutputBuffer, UpstreamOutputBuffer, DownstreamOutputBuffer, DebugOutputBuffer;
	int oldw = 0, oldh = 0;

#ifdef _WIN32
	{
		struct stat filestat;
		if (stat("stripclub.exe", &filestat)) {	// Can't find the executable, so the working dir must be wrong (sanity check)
			MOut("Strip Club should not be run from any directory but the location of the executable. Working directory is pointing to the wrong place, perhaps?\n");
			return 1;
		}
	}
#endif

	if(SDL_Init(0)==-1) {
		Out("SDL_Init: %s\n", SDL_GetError());
		exit(1);
	}
	if(SDLNet_Init()==-1) {
		Out("SDLNet_Init: %s\n", SDLNet_GetError());
		exit(2);
	}

	CurComic = "debug";			// Dirty hack, but hey...
	debug = fopen(GetCacheFileName("debug.log"), "w");
//	debug = stdout;

	Quit = false;

	CurComic = CurLinks.FirstLink = CurLinks.PrevLink = CurLinks.CurLink = CurLinks.NextLink = CurLinks.LatestLink = NULL;

	InitOutputWindow();
	MakeMiscWindows();
	Prefs = new Fl_Preferences(Fl_Preferences::USER, "racercksproductions", "stripclub");

	TextOutput->buffer(&TextOutputBuffer);
	UpstreamOutput->buffer(&UpstreamOutputBuffer);
	DownstreamOutput->buffer(&DownstreamOutputBuffer);
	DebugOutput->buffer(&DebugOutputBuffer);

	strcpy(WindowTitle, SCVERSION);

	Interface.Make();

	Interface.MainWindow->show();

	{
		int remx, remy, remw, remh;
		Prefs->get("remx", remx, 4);
		Prefs->get("remy", remy, 4);
		Prefs->get("remw", remw, 640);
		Prefs->get("remh", remh, 400);
		Interface.MainWindow->resize(remx, remy, remw, remh);
	}

	{
		int tmp;
		Prefs->get("offline", tmp, OFFTYPE_MISSING);
		Offline = (eOffType)tmp;
		Interface.OutputOffline->value(OffArray[Offline]);
	}

	if (!Prefs->entryExists("warning")) {
		SplashWarning->show();
		Prefs->set("warning", 1);
	}

	InitMenu();

	while (Interface.MainWindow->visible() && !Quit) {
		if ((oldw != Interface.MainWindow->w()) || (oldh != Interface.MainWindow->h())) {
			oldw = Interface.MainWindow->w();
			oldh = Interface.MainWindow->h();
			Interface.MainMenu->resize(0, 0, 144, 20);
			Interface.OutputURL->resize(144, 0, oldw - 208, 20);
			Interface.OutputOffline->resize(oldw - 64, 0, 64, 20);
			Interface.ScrollBox->resize(2, 22, oldw - 4, oldh - 46);
			Interface.Progress->resize(2, oldh - 22, oldw - 288, 20);
			Interface.FirstButton->resize(oldw - 284, oldh - 22, 56, 20);
			Interface.PrevButton->resize(oldw - 228, oldh - 22, 56, 20);
			Interface.RefreshButton->resize(oldw - 172, oldh - 22, 56, 20);
			Interface.NextButton->resize(oldw - 116, oldh - 22, 56, 20);
			Interface.LatestButton->resize(oldw - 60, oldh - 22, 56, 20);
			Interface.MainWindow->redraw();
		}
		if (!OutputWindow->visible()) {
			OutputWindow->show();
		}
		Fl::wait();
	}

	SDLNet_Quit();

	SDL_Quit();

	if(CurComicG) {
		delete CurComicG;
	}

	Prefs->set("remx", Interface.MainWindow->x());
	Prefs->set("remy", Interface.MainWindow->y());
	Prefs->set("remw", Interface.MainWindow->w());
	Prefs->set("remh", Interface.MainWindow->h());
	Prefs->set("offline", (int)Offline);

	delete Interface.MainWindow;
	delete OutputWindow;
	delete Prefs;

	fclose(debug);

	remove("temp.html");
	remove("temp.txt");

	return 0;
}

void InitMenu() {
	Fl_Preferences *TempGroup = NULL;
	int i, comics = Prefs->groups();
	static char **groupnames = NULL;		// Static because the pointers are needed for the callback
	char buf[160], *name, *category;
	Interface.MainMenu->menu(cWindow::menu_MainMenu); 		// Reset the menu in case we're reloading
	if(!comics) {
		MOut("No comics found. Make sure you import/create some.\n");
		if (fl_ask("Run updater to download .comic files off the Strip Club site?")) {
			UpdateComics();
			MOut("Comic files can be found in %s\nMake sure to import some!\n", GetComicFileName(""));
		}
		return;
	}
	if (groupnames) {
		for (i = 0; groupnames[i]; i++) {
			delete groupnames[i];
		}
		delete groupnames;
	}
	groupnames = new char*[comics + 1];
	groupnames[comics] = 0;			// To mark the end of the array
	for (i = 0; i < comics; i++) {
		clonestr(Prefs->group(i), (groupnames + i));
		if(strlen(groupnames[i])) {
			Out("Loading Comic: %s\n", groupnames[i]);
			if (TempGroup) {
				delete TempGroup;
			}
			TempGroup = new Fl_Preferences(Prefs, groupnames[i]);
			if (TempGroup->entryExists("name")) {
				TempGroup->get("name", name, "");
				TempGroup->get("category", category, "Unsorted");
				sprintf(buf, "Comics/%s/%s/Read", category, name);
				Interface.MainMenu->add(buf, 0, SwitchComic, groupnames[i]);
				sprintf(buf, "Comics/%s/%s/Edit", category, name);
				Interface.MainMenu->add(buf, 0, EditComic, groupnames[i]);
				sprintf(buf, "Comics/%s/%s/Export", category, name);
				Interface.MainMenu->add(buf, 0, ExportComic, groupnames[i]);
				delete name;
				delete category;
			}
		}
	}
	if (TempGroup) {
		delete TempGroup;
	}
}

void AddComic() {
	const char *Buffer;
	char *GroupName;
	Buffer = fl_input("Enter a short name for the new comic (eg: sluggy)\n", "comic");
	if (!Buffer || !strlen(Buffer)) {
		return;
	}
	clonestr(Buffer, &GroupName);
	if (Prefs->groupExists(GroupName)) {
		if (!fl_ask("Comic already exists! Edit?")) {
			delete GroupName;
			return;
		}
	}
	EditComic(0, GroupName);
	delete GroupName;
}

void EditComic(Fl_Widget *w, void *Comic) {
	Fl_Preferences *EditComic = new Fl_Preferences(Prefs, (char *)Comic);
	char Buffer[400];
	EditComic->get("name", Buffer, "New Comic", 399);
	InputName->value(Buffer);
	EditComic->get("baseurl", Buffer, "http://", 399);
	InputBaseURL->value(Buffer);
	EditComic->get("firsturl", Buffer, "http://", 399);
	InputFirstURL->value(Buffer);
	EditComic->get("imgpattern", Buffer, "<img .*src=\"(.*comics/.*(png|jpg|gif))\".*>", 399);
	InputImgPattern->value(Buffer);
	EditComic->get("linktype", Buffer, "tag", 399);
	if (!strcmp("tag", Buffer)) {
		ChoiceLinkType->value(0);
	} else if (!strcmp("text", Buffer)) {
		ChoiceLinkType->value(1);
	}	
	EditComic->get("prevpattern", Buffer, "<img .*name=\"previous_day\".*>", 399);
	InputPrevLink->value(Buffer);
	EditComic->get("nextpattern", Buffer, "<img .*name=\"next_day\".*>", 399);
	InputNextLink->value(Buffer);
	EditComic->get("relativeurl", Buffer, "", 399);
	InputRelative->value(Buffer);
	if (EditComic->entryExists("useragent")) {
		CheckDefaultUserAgent->value(0);
		EditComic->get("useragent", Buffer, "", 399);
		InputUserAgent->value(Buffer);
		InputUserAgent->activate();
		ChoiceUserAgent->activate();
	} else {
		CheckDefaultUserAgent->value(1);
		InputUserAgent->value(USERAGENT);
		InputUserAgent->deactivate();
		ChoiceUserAgent->deactivate();
	}
	EditComic->get("category", Buffer, "Unsorted", 399);
	InputCategory->value(Buffer);
	EditComicWindow->show();
	while (EditComicWindow->visible()) { Fl::wait(); }
	if (!strcmp("delete", InputName->value())) {
		if (EditComic->entryExists("name")) {		// Comic exists, so ask if we want to delete the associated files
			if (fl_ask("Delete locally saved files?")) {
				char cmd[1024];
				CurComic = (char *)Comic;
				Out("Removing files for %s...\n", (char *)Comic);
#ifdef _WIN32
				// ".\" because the LAST thing we want this thing to do is delete shit in the root directory if it messes up
				snprintf(cmd, 1023, "del .\\%s\\*.*", (char *)Comic);	
				system(cmd);
				snprintf(cmd, 1023, "rmdir .\\%s", (char *)Comic);
				system(cmd);
#else
				snprintf(cmd, 1023, "rm -rf %s", GetCacheFileName("", false));
				system(cmd);
#endif
				CurComic = NULL;
			}
		}
		delete EditComic;
		EditComic = NULL;
		Prefs->deleteGroup((char *)Comic);
	} else if (strcmp("cancel", InputName->value())) {		// We did NOT click "Cancel"
		EditComic->set("name", InputName->value());
		EditComic->set("baseurl", InputBaseURL->value());
		EditComic->set("firsturl", InputFirstURL->value());
		EditComic->set("imgpattern", InputImgPattern->value());
		switch(ChoiceLinkType->value()) {
			case(0): {
				EditComic->set("linktype", "tag");
				break;
			}
			case(1): {
				EditComic->set("linktype", "text");
				break;
			}
		}
		EditComic->set("prevpattern", InputPrevLink->value());
		EditComic->set("nextpattern", InputNextLink->value());
		EditComic->set("relativeurl", InputRelative->value());
		if (CheckDefaultUserAgent->value()) {		// Default is checked
			EditComic->deleteEntry("useragent");
		} else {									// Not checked
			EditComic->set("useragent", InputUserAgent->value());
		}
		EditComic->set("category", InputCategory->value());
	}
	if (EditComic) {
		if (!EditComic->entryExists("name")) {		// New comic, but cancel was clicked, so make sure it's not stored
			delete EditComic;
			Prefs->deleteGroup((char *)Comic);
		} else {
			delete EditComic;
		}
	}
	InitMenu();
	Prefs->flush(); 		// Make sure changes are written to disk
}

extern IPaddress ip;	// Holds host info for current comic

void SetOffline(int i)
{
	if (Offline == OFFTYPE_OFFLINE) {		// Look up the host if switching out of offline mode
		sURL Break;
		char *URL;
		CurComicG->get("baseurl", URL, "");
		URLBreak(URL, &Break);
		DbgOut("Looking up host: %s\n", Break.Host);
		Interface.BarLabel("Looking up host...");

		if (SDLNet_ResolveHost(&ip, Break.Host, atoi(Break.Port))) {
			Interface.BarLabel("Unable to resolve host!");
			MOut("Unable to resolve host: %s\n", Break.Host);
			delete URL;
			URLDelete(&Break);
			return;
		}
		delete URL;
		URLDelete(&Break);
	}
	Offline = (eOffType)i;
	Interface.OutputOffline->value(OffArray[Offline]);
}
